import ctypes as c
from PySide.QtCore import *
''' 
 *
 *  Copyright (C) 2013  André Luís da Costa, FEEC, Unicamp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
'''

from PySide.QtGui import *
import numpy as np
from scipy.interpolate import UnivariateSpline
from time import time

try:
	import OpenGL
	import OpenGL.GL as GL
except ImportError:
	app = QApplication(sys.argv)
	QMessageBox.critical(None, "Visualizador de volumes",
						"PyOpenGL must be installed to run this example.",
						QMessageBox.Ok | QMessageBox.Default,
						QMessageBox.NoButton)
	sys.exit(1)

class Volume:
	def __init__(self, nifti_image):
		self._data = nifti_image.get_data()
		umin, umax = self._data.min(), self._data.max()
		self._data = np.round(255*(self._data - umin)/(umax-umin)).astype('uint8')
		self._vox_dim = nifti_image.get_header()['pixdim'][1:4]
		T = nifti_image.get_header().get_sform()
		c = np.array(
			[[1,0,0],
			 [0,1,0],
			 [0,0,1],
			 [0,0,0]])
		c = np.dot(T,c)[:3]
		t = np.argmax(abs(c), axis=1)
		s = c[np.arange(3), t][np.argsort(t)]
		s = np.round(s/abs(s)).astype('int8')
		self._data = self._data[::s[0], ::s[1], ::s[2]]
		self._data = self._data.transpose(t).copy()
		self._vox_dim = self._vox_dim[t]
		self._cube = np.array(
			[[0, 0, 0],
			 [0, 1, 0],
			 [1, 1, 0],
			 [1, 0, 0],
			 
			 [0, 0, 0],
			 [1, 0, 0],
			 [1, 0, 1],
			 [0, 0, 1],
			 
			 [0, 0, 0],
			 [0, 0, 1],
			 [0, 1, 1],
			 [0, 1, 0],
			 
			 [0, 0, 1],
			 [1, 0, 1],
			 [1, 1, 1],
			 [0, 1, 1],
			 
			 [1, 0, 0],
			 [1, 1, 0],
			 [1, 1, 1],
			 [1, 0, 1],
			 
			 [0, 1, 0],
			 [0, 1, 1],
			 [1, 1, 1],
			 [1, 1, 0]], dtype='float32')
		self.__make_parallelepiped()
	
	def __make_parallelepiped(self):
		self._shape = np.array(self._data.shape, dtype='int')
		s = self.get_shape(True)
		self._parallelepiped = (s * self._cube).astype('float32')
	
	def get_parallelepiped(self):
		return self._parallelepiped
	
	def get_texture_coordinates(self):
		return self._cube
	
	def get_data(self):
		return self._data
	
	def get_shape(self, normed=False):
		if normed:
			size = self._vox_dim*self._shape
			return size/size.max()
		else:
			return self._shape

class Spline_Knot():
	def __init__(self, x, y, x_locked=False, erasable=True):
		self._x_locked = x_locked
		self._erasable = erasable
		self._origin = (x,y)
		self.set_position(x,y)
		self._radius = 0.012
	
	def draw(self):
		x,y = self.get_position()
		GL.glColor3f(0.8,1,0.5)
		GL.glLineWidth(2)
		GL.glBegin(GL.GL_LINE_STRIP)
		T = np.linspace(-np.pi,np.pi,10)
		for t in T:
			GL.glVertex3f(x + np.cos(t)*self._radius, y + np.sin(t)*self._radius, 0)
		GL.glEnd()
	
	def set_position(self, x, y):
		x = min(1, max(0, x))
		y = min(1, max(0, y))
		if self._x_locked:
			self._position = (self._origin[0], y)
		else:
			self._position = (x,y)
	
	def get_position(self):
		return self._position
	
	def pick(self, x, y):
		point = np.array([x,y])
		corner_1 = np.array(self._position) - 2*self._radius
		corner_2 = np.array(self._position) + 2*self._radius
		return np.logical_and(point>=corner_1, point<=corner_2).all()

class Opacity_Function():
	def __init__(self):
		self._knots = []
		self.__add_knot(0,0, True, False)
		self.__add_knot(0.5,0.5)
		self.__add_knot(1,1, True, False)
		self.__gen_spline()
		self._current_knot = None
	
	def __add_knot(self, x, y, x_locked=False, erasable=True):
		self._knots.append(Spline_Knot(x, y, x_locked, erasable))
	
	def __gen_spline(self):
		X, Y = [], []
		for knot in self._knots:
			x,y = knot.get_position()
			X.append(x)
			Y.append(y)
		X, Y = np.array(X), np.array(Y)
		seq = np.argsort(X)
		X, Y = X[seq], Y[seq]
		self.spline = UnivariateSpline(X, Y, k=2, s=0)		
	
	def __draw_background(self):
		# Desenha o fundo do grafico de opacidade
		GL.glColor3f(0.1,0.07,0)
		GL.glBegin(GL.GL_QUADS)
		GL.glVertex3f(0,0,0)
		GL.glVertex3f(1,0,0)
		GL.glVertex3f(1,1,0)
		GL.glVertex3f(0,1,0)
		GL.glEnd()
		
		# Desenha a grade do grafico de opacidade
		tics = np.arange(0.2,1,0.2)
		GL.glColor3f(0.2,0.18,0.0)
		GL.glLineStipple(1, 0xF0F0)
		GL.glEnable(GL.GL_LINE_STIPPLE)
		GL.glLineWidth(1)
		GL.glBegin(GL.GL_LINES)
		for tic in tics:
			GL.glVertex3f(tic,0,0)
			GL.glVertex3f(tic,1,0)
			GL.glVertex3f(0,tic,0)
			GL.glVertex3f(1,tic,0)		
		GL.glEnd()
	
	def __draw_frame(self):
		# Desenha a borda do grafico de opacidade
		GL.glColor3f(0.7,0.65,0.0)
		GL.glDisable(GL.GL_LINE_STIPPLE)
		GL.glLineWidth(1.5)
		GL.glBegin(GL.GL_LINES)
		for tic in [0,1]:
			GL.glVertex3f(tic,0,0)
			GL.glVertex3f(tic,1,0)
			GL.glVertex3f(0,tic,0)
			GL.glVertex3f(1,tic,0)		
		GL.glEnd()
	
	def __draw_curve(self):
		GL.glColor3f(0,0.9,0.2)
		GL.glLineWidth(1.5)
		GL.glBegin(GL.GL_LINE_STRIP)
		X = np.linspace(0,1,150)
		for x in X:
			GL.glVertex3f(x,min(1,max(0, self.spline(x))),0)
		GL.glEnd()
	
	def __draw_knots(self):
		for knot in self._knots:
			knot.draw()
	
	def draw(self):
		self.__draw_background()
		self.__draw_frame()
		self.__draw_curve()
		self.__draw_knots()
	
	def erase_knot(self, x, y):
		for knot in self._knots[::-1]:
			if knot.pick(x,y):
				if knot._erasable:
					if self._current_knot==knot:
						self._current_knot = None
					self._knots.remove(knot)
					if len(self._knots)<3:
						self.__add_knot(0.5,0.5)
				else:
					knot.set_position(*knot._origin)
				self.__gen_spline()
				return True
		return False
	
	def create_knot(self, x, y):
		pos = np.array([x,y])
		if not np.logical_or(pos<0, pos>1).any():
			self.__add_knot(*pos)
			self.__gen_spline()
			return True
		return False
	
	def select_knot(self, x, y):
		self._current_knot = None
		for knot in self._knots[::-1]:
			if knot.pick(x,y):
				self._current_knot = knot
				return True
		return False
	
	def move_selected_knot(self, x, y):
		if not self._current_knot is None:
			self._current_knot.set_position(x, y)
			self.__gen_spline()
	
	def get(self, n):
		x = np.linspace(0,1,n)
		return np.minimum(1, np.maximum(0, self.spline(x)))

class Color_Knot():
	_y = 1.2
	def __init__(self, x, color=(1,1,1), erasable=True):
		self.set_color(color)
		self._erasable = erasable
		self._origin = x
		self.set_position(x)
		self._base = 0.012
	
	def draw(self, selected=False):
		if selected:
			GL.glColor3f(0.8,0.1,0.6)
		else:
			GL.glColor3f(0.1,0.8,0.6)
		GL.glBegin(GL.GL_TRIANGLES)
		GL.glVertex3f(self.get_position() - self._base, self._y,0)
		GL.glVertex3f(self.get_position() + self._base, self._y,0)
		GL.glVertex3f(self.get_position(), self._y + 3*self._base,0)
		GL.glEnd()
		GL.glBegin(GL.GL_QUADS)
		GL.glVertex3f(self.get_position() - self._base, self._y,0)
		GL.glVertex3f(self.get_position() - self._base, self._y - 4*self._base,0)
		GL.glVertex3f(self.get_position() + self._base, self._y - 4*self._base,0)
		GL.glVertex3f(self.get_position() + self._base, self._y,0)
		GL.glEnd()
	
	def set_position(self, x):
		if self._erasable:
			x = min(1, max(0, x))
			self._position = x
	
	def get_position(self):
		if self._erasable:
			return self._position
		else:
			return self._origin
	
	def set_color(self, color):
		self._color = color
	
	def get_color(self):
		return self._color
	
	def pick(self, x, y):
		return x>=(self.get_position()-self._base) and x<=(self.get_position()+self._base)

class Color_Function():
	def __init__(self):
		self._knots = []
		self.__add_knot(0, (0,0,0), False)
		self.__add_knot(1, (1,1,1), False)
		self._current_knot = None
		self._TF = None
	
	def __add_knot(self, x, color=(1,1,1), erasable=True):
		new_knot = Color_Knot(x, color, erasable)
		self._knots.append(new_knot)
		return new_knot
	
	def __draw_background(self):
		n = len(self._knots)
		seq = np.argsort([self._knots[x].get_position() for x in xrange(n)])
		seq = seq[np.round(np.linspace(0,n-1,2*n-2)).astype('int')]
		colors = []
		y0,y1 = 1.5,1.2
		for i in seq:
			x = self._knots[i].get_position()
			r,g,b = self._knots[i].get_color()
			GL.glColor3f(r,g,b)
			if y0==1.5:
				GL.glBegin(GL.GL_QUADS)
			
			GL.glVertex3f(x, y0, 0)
			GL.glVertex3f(x, y1, 0)
			
			if y0==1.2:
				GL.glEnd()
			y0,y1 = y1,y0
	
	def __draw_frame(self):
		# Desenha a borda da barra de cores
		GL.glColor3f(0.7,0.65,0.0)
		GL.glDisable(GL.GL_LINE_STIPPLE)
		GL.glLineWidth(1.0)
		GL.glBegin(GL.GL_LINES)
		for tic in [1.2,1.5]:
			GL.glVertex3f(0,tic,0)
			GL.glVertex3f(1,tic,0)		
		for tic in [0,1]:
			GL.glVertex3f(tic, 1.2, 0)
			GL.glVertex3f(tic, 1.5, 0)		
		GL.glEnd()
	
	def __draw_knots(self):
		for knot in self._knots:
			knot.draw(self._current_knot==knot)
	
	def draw(self):
		self.__draw_background()
		self.__draw_frame()
		self.__draw_knots()
	
	def __erase_selected_knot(self):
		if not self._current_knot is None:
			if self._current_knot._erasable:
				self._knots.remove(self._current_knot)
				self._current_knot = None
				self._TF = None
			else:
				return False
		return True
	
	def erase_knot(self, x, y):
		self.select_knot(x, y)
		return self.__erase_selected_knot()
	
	def create_knot(self, x, y):
		if x>=0 and x<=1:
			tf = self.get(256)
			i = int(np.round(x*255))
			self._current_knot = self.__add_knot(x, tf[i])
			self.select_color()
			self._TF = None
			return True
		return False
	
	def select_color(self):
		if not self._current_knot is None:
			color = QColorDialog.getColor(initial= QColor.fromRgbF(*self._current_knot.get_color()), options=QColorDialog.NoButtons)
			if color.isValid():
				self._current_knot.set_color(np.array(color.getRgb(), 'float32')[:3]/255.0)
				self._TF = None
				return True
		return False
	
	def select_knot(self, x, y):
		self._current_knot = None
		for knot in self._knots[::-1]:
			if knot.pick(x,y):
				self._current_knot = knot
				return True
		return False
	
	def move_selected_knot(self, x, y):
		if not self._current_knot is None:
			self._current_knot.set_position(x)
			self._TF = None
	
	def get(self, n):
		if self._TF is None:
			tf = [[],[],[]]
			
			m = len(self._knots)
			seq = np.argsort([self._knots[x].get_position() for x in xrange(m)])
			seq = seq[np.round(np.linspace(0,m-1,2*m-2)).astype('int')]
			for i in xrange(0,len(seq),2):
				aknot = self._knots[seq[i]]
				bknot = self._knots[seq[i+1]]
				x0,x1 = aknot.get_position(), bknot.get_position()
				acolor, bcolor = aknot.get_color(), bknot.get_color()
				N = int(np.round(x1*n) - np.round(x0*n))
				x = np.linspace(0,1,N)
				for c in xrange(3):
					tf[c].extend(acolor[c]*(1-x) + bcolor[c]*x)
			self._TF = np.array(tf).T
		return self._TF

class Maxtree:
	def __init__(self, volume):
		self._volume = volume
		self._builder = c.CDLL("maxtree/libmaxtree.so")
		self.__build_tree()
	
	def __build_tree(self):
		data = self._volume.ravel()
		n = data.size
		t = time()
		self._belongs_to = np.empty(n, dtype='uint32')
		self._parent = np.ones(n, dtype='uint32')*n
		print 'Support vectors', time()-t
		self._builder.createMaxtree(self._volume.ctypes.data_as(c.POINTER(c.c_ubyte)),
									self._volume.ctypes.shape_as(c.c_uint), 3,
									self._belongs_to.ctypes.data_as(c.POINTER(c.c_uint)),
									self._parent.ctypes.data_as(c.POINTER(c.c_uint)))
		print 'Maxtree built', time()-t
		H = np.bincount(self._belongs_to.astype('int32'))
		T = {}
		self._root = (np.arange(n)==self._parent).nonzero()[0][0]
		idx = self._parent!=n
		self._nodes = idx.nonzero()[0]
		leaf = np.ones(n, dtype='bool')
		leaf[self._parent[idx]] = False
		leaves = self._nodes[leaf[idx]]
		visited = np.zeros(n, dtype='bool')
		for i in leaves:
			j = i
			k = None
			ac = 0
			while True:
				if visited[j]:
					T[j][1] += ac
				else:
					ac += H[j]
					T[j] = [self._parent[j], ac, data[j], data[j], []]
				if k is not None:
					T[j][4].append(k)
					k = None
				if not visited[j]:
					visited[j] = True
					k = j
				if j==self._root:
					break
				j = self._parent[j]
		
		print 'Tree recover', time()-t
		self.T = T
	
	def reset_filters(self):
		q = [self._root]
		while len(q)>0:
			p = q.pop(0)
			self.T[p][3] = self.T[p][2]
			q.extend(self.T[p][4])
	
	def volume_filter(self, t1, t2):
		n = 0
		q = [self._root]
		while len(q)>0:
			p = q.pop(0)
			if self.T[p][1]<t1 or self.T[p][1]>t2:
				self.T[p][3] = self.T[self.T[p][0]][3]
			else:
				self.T[p][3] = self.T[p][2]
				n += 1
			q.extend(self.T[p][4])
		return n
		
	def get_filtered_data(self):
		lvls = np.zeros(self._volume.size, dtype='uint8')
		for k,v in self.T.items():
			lvls[k] = v[3]
		return lvls[self._belongs_to].reshape(self._volume.shape)
